;nyquist plug-in
;version 4
;type process
;preview enabled
;name "Sinc Filter"
;action "Applying Filter..."
;author "Steve Daulton"
;release 2.4.0
;copyright "Released under terms of the GNU General Public License version 2"

;control type "Filter Type" choice "Low Pass,High Pass" 0
;control hz "Filter frequency" float "Hz" 1000 1 20000
;control bandwidth "Bandwidth" float "Hz" 100 1 1000


;; Bandwidth ratio as a fraction of the sampling rate
(setf bandwidth (/ bandwidth *sound-srate*))

;; Frequency as a fraction of sample rate
(setf fc (/ hz *sound-srate*))

;; Calculate kernel length (must be even)
;; http://www.dspguide.com/ch16/2.htm
(setf klength (/ 4.0 bandwidth))
(setf halfk (round (/ klength 2)))
(setf klength (* 2 halfk))


(defun error-check ()
  (when (< *sound-srate* 10)
    (throw 'err "Error.~%Track sample rate below 10 Hz is not supported."))
  (when (> hz (* 0.495 *sound-srate*))
    (throw 'err
        (format nil "Error.~%\"Filter frequency\" (~a Hz) is too high for selected track.~%~
                Set Filter frequency below ~a Hz."
            hz
            (truncate (* 0.495 *sound-srate*))))))

(defun sinc (x fc)
  ;; http://www.dspguide.com/ch16/1.htm
  ;; Note that fc is a fraction of the sample rate.
  (if (= x 0)
      (* 2 pi fc)
      (/ (sin (* 2 pi fc x)) x)))

(defun blackman (i M)
  ;; Where: 0 <= i <= M
  (+ 0.42
     (* -0.5 (cos (/ (* 2.0 pi i) M)))
     (* 0.08 (cos (/ (* 4 pi i) M)))))


(defun calc-kernel (size fc)
  ;; Generate windowed sinc kernel impulse
  (when (oddp size)
    (error "Size of sinc filter must be even"))
  (let ((ar (make-array (1+ size)))
        (norm 0)  ;Normalization factor
        val)
    (do ((i 0 (1+ i))
         (j size (1- j))
         (x (- halfk) (1+ x)))
        ((> i j))
      (setf val (* (sinc x fc)(blackman i size)))
      (setf norm (+ norm val))
      (setf (aref ar i) val)
      (setf (aref ar j) val))
    ;; norm is sum of all samples, but don't count middle value twice.
    (setf norm (- (* norm 2)(aref ar halfk)))
    (dotimes (i size ar)
      (setf (aref ar i)(/ (aref ar i) norm)))))

(defun get-kernel (size fc type)
  ;; type: 0 (low pass) or 1 (highpass)
  ;; Returns filter kernel as a sound.
  (let ((kernel (calc-kernel size fc)))
    (when (= type 1)
      ;; Convert kernel to high pass
      ;; https://tomroelandts.com/articles/how-to-create-a-simple-high-pass-filter
      (dotimes (i size kernel)
        (setf (aref kernel i)(* -1 (aref kernel i))))
      (incf (aref kernel halfk)))
    (snd-from-array 0 *sound-srate* kernel)))

(defun sinc-filter (sig start end impulse)
  (extract-abs start end (convolve sig impulse)))

(defun process ()
  (error-check)
  (let ((imp (get-kernel klength fc type))
        (start (/ halfk *sound-srate*))
        (dur (get-duration 1)))
    (multichan-expand #'sinc-filter *track* start (+ start dur) imp)))

(catch 'err (process))
